// https://doc.rust-lang.org/src/std/io/mod.rs.html

use crate::CifError;
use arrayvec::{ArrayString, ArrayVec};
use std::io::{Error, Read};

struct Guard<'a> {
    buf: &'a mut ArrayVec<u8, 8192>,
    len: usize,
}

impl Drop for Guard<'_> {
    fn drop(&mut self) {
        unsafe {
            self.buf.set_len(self.len);
        }
    }
}

pub trait UnsafeBufRead: Read {
    fn consume(&mut self, amt: usize);
    fn fill_buf(&mut self) -> Result<&[u8], Error>;
    fn read_until(&mut self, byte: u8, buf: &mut ArrayVec<u8, 8192>) -> Result<usize, CifError> {
        read_until(self, byte, buf)
    }

    fn read_line(&mut self, buf: &mut ArrayString<8192>) -> usize {
        let mut buf_vec = ArrayVec::<u8, 8192>::new();
        for u in buf.as_bytes() {
            buf_vec.push(*u);
        }
        let return_val = read_until(self, b'\n', &mut buf_vec).unwrap();
        unsafe { buf.push_str(std::str::from_utf8_unchecked(buf_vec.as_slice())) };
        return_val
    }

    fn lines(self) -> UnsafeLines<Self>
    where
        Self: Sized,
    {
        UnsafeLines { buf: self }
    }
}

#[derive(Debug)]
pub struct UnsafeLines<B> {
    buf: B,
}

impl<B: UnsafeBufRead> Iterator for UnsafeLines<B> {
    type Item = Result<ArrayString<8192>, CifError>;

    fn next(&mut self) -> Option<Result<ArrayString<8192>, CifError>> {
        let mut buf = ArrayString::<8192>::new();
        match self.buf.read_line(&mut buf) {
            0 => None,
            _n => {
                if buf.ends_with('\n') {
                    buf.pop();
                }
                Some(Ok(buf))
            }
        }
    }
}

fn read_until<R: UnsafeBufRead + ?Sized>(r: &mut R, delim: u8, buf: &mut ArrayVec<u8, 8192>) -> Result<usize, CifError> {
    let mut read = 0;
    loop {
        let (done, used) = {
            let available = match r.fill_buf() {
                Ok(n) => n,
                Err(ref e) if e.kind() == std::io::ErrorKind::Interrupted => continue,
                Err(e) => return Err(CifError::IoError(e)),
            };
            match core::slice::memchr::memchr(delim, available) {
                Some(i) => {
                    buf.try_extend_from_slice(&available[..=i])?;
                    (true, i + 1)
                }
                None => {
                    buf.try_extend_from_slice(available)?;
                    (false, available.len())
                }
            }
        };
        r.consume(used);
        read += used;
        if done || used == 0 {
            return Ok(read)
        }
    }
}
